{**************************************}
{* Main unit for Bacterio application *}
{**************************************}

unit bacterio;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, Menus, StdCtrls,
  ExtCtrls;

type
  TBacteria = record
    Name, Stain, Shape, Wall, Growth: string;
  end;
  TBacterias = array[1..23] of TBacteria;
  TInfection = record
    ScenarioGroup, ScenarioActual: string;
    ScenarioBacterias: array[1..23] of Char;
  end;
  TInfections = array of TInfection;
  TSuppl = record
    ScenarioUncommon, ScenarioBacteria, ScenarioBacteriaDetails: string;
  end;
  TSuppls = array[1..9] of TSuppl;
  {************}
  { TfBacterio }
  {************}
  TfBacterio = class(TForm)
    mMenu: TMainMenu;
    mFile, mFileExit: TMenuItem;
    mOptions, mOptionsId, mHelp, mHelpAbout: TMenuItem;
    mOptionsId1, mOptionsId2, mOptionsId3, mOptionsId4: TMenuItem;
    StaticText1: TStaticText;
    Memo1: TMemo;
    Label1, Label2, Label3, Label4, Label5: TLabel;
    cobSelection, cobActual, cobActual2: TComboBox;
    rgStain, rgWall, rgShape, rgGrowth: TRadioGroup;
    edCauses: TMemo;
    edCause: TEdit;
    btSearch: TButton;
    procedure FormCreate(Sender: TObject);
    procedure mFileExitClick(Sender: TObject);
    procedure mOptionsId1Click(Sender: TObject);
    procedure mOptionsId2Click(Sender: TObject);
    procedure mOptionsId3Click(Sender: TObject);
    procedure mOptionsId4Click(Sender: TObject);
    procedure mHelpAboutClick(Sender: TObject);
    procedure btSearchClick(Sender: TObject);
    procedure cobSelectionChange(Sender: TObject);
    procedure cobActualChange(Sender: TObject);
  private
    iIdentification: Integer;
    aBacterias: TBacterias;
    aInfections: TInfections;
    aSuppls: TSuppls;
  end;

var
  fBacterio: TfBacterio;

implementation

{$R *.lfm}

{ Read program data from text files }

procedure ReadData(out Bacterias: TBacterias; out Infections: TInfections; out Suppls: TSuppls);

var
  N, G, B, I, P: Integer;
  Line, S: string;
  InFile: Text;

begin
  // Bacteria data
  Assign(InFile, 'bacteria.txt'); Reset(InFile); N := 0;
  while not EoF(InFile) do begin
    Readln(Infile, Line); Line := Trim(Line);
    if Line <> '' then begin
      Inc(N);
      with Bacterias[N] do begin
        // Read bacteria name, stain, shape, growth condition
        Name := Trim(LeftStr(Line, 45));                                       // cols  1-45: Name
        S := Copy(Line, 46, 2); Stain := '';                                   // cols 46-47: Stain
        if S = 'G+' then
          Stain := 'Gram-positive'
        else if S = 'G-' then
          Stain := 'Gram-negative'
        else if S = 'AF' then
          Stain := 'Acid-fast';
        Shape := Trim(Copy(Line, 51, 10));                                     // cols 51-60: Shape
        S := Copy(Line, 61, 3); Growth := '';                                  // cols 61-63: Growth condition
        if S = 'AER' then
          Growth := 'Aerobical'
        else if S = 'ANA' then
          Growth := 'Anaerobical'
        else if S = 'FAN' then
          Growth := 'Facultative anaerobe'
        else if S = 'NCU' then
          Growth := 'Non culturable';
        // Determine bacteria cell wall
        Wall := '';
        if Growth = 'Non culturable' then
          Wall := 'None'
        else if Stain = 'Gram-positive' then
          Wall := 'Thick'
        else if Stain = 'Gram-negative' then
          Wall := 'Thin + membrane'
        else if Stain = 'Acid-fast' then
          Wall := 'Contains mycolic acid';
      end;
    end;
  end;
  Close(InFile);
  // Clinical scenarios (infections) data
  Assign(InFile, 'infections.txt'); Reset(InFile); N := 0;
  while not EoF(InFile) do begin
    Readln(Infile, Line); Line := Trim(Line);
    if Line <> '' then begin
      Inc(N); SetLength(Infections, N);
      with Infections[N - 1] do begin
        // Scenario group and scenario name
        G := StrToInt(LeftStr(Line, 1));                                       // cols 1-1:  group index
        ScenarioGroup := fBacterio.cobSelection.Items[G - 1];                  // scenario group name taken from combobox
        ScenarioActual := Copy(Line, 4, 40); B := 0;                           // cols 4-43: scenario name
        // Bacterial causes for this scenario
        for I := 44 to 88 do begin
          // For each of the 23 bacteria, read code for this scenario
          if I mod 2 = 0 then begin
            Inc(B);
            ScenarioBacterias[B] := Copy(Line, I, 1)[1];                       // codes: x = yes, - = no, n = note index (suppl.txt)
            if ScenarioBacterias[B] = '?' then                                 // knowledge lack of myself; consider being "no"
              ScenarioBacterias[B] := '-';
          end;
        end;
      end;
    end;
  end;
  Close(InFile);
  // Supplementary data
  Assign(InFile, 'suppl.txt'); Reset(InFile); N := 0;
  while not EoF(InFile) do begin
    Readln(Infile, Line); Line := Trim(Line);
    if Line <> '' then begin
      Inc(N);
      with Suppls[N] do begin
        ScenarioUncommon := Trim(Copy(Line, 1, 65));                           // cols  1-65: Uncommon causes for given infection under certain conditions
        if ScenarioUncommon = '---' then
          ScenarioUncommon := '';
        ScenarioBacteria := Trim(Copy(Line, 66, Length(Line)));                // cols 66-  : Common bacteria for given infection under certain conditions
        if ScenarioBacteria = '---' then begin
          ScenarioBacteria := ''; ScenarioBacteriaDetails := '';
        end
        else begin
          // Common bacteria for given infection under certain conditions:
          // Can give a more precise bacterial cause; e.g. for neonatal meningitis
          // Common cause is Beta-haemolytic Streptococcus, group B (out of Beta-haemolytic Streptococcus, group A, B, C, or G)
          P := Pos(':', ScenarioBacteria);                                     // "general" bacteria cause and precise cause separated by colon
          ScenarioBacteriaDetails := Copy(ScenarioBacteria, P + 2, Length(ScenarioBacteria));
          ScenarioBacteria := LeftStr(ScenarioBacteria, P - 1);
        end;
      end;
    end;
  end;
  Close(InFile);
end;

{ Fill scenario combobox (for given scenario group) }

procedure FillScenarios(ScenarioGroup: string; var Infections: TInfections);

var
  I: Integer;

begin
  fBacterio.cobActual.Items.Clear;
  for I := 0 to Length(Infections) - 1 do begin
    if ScenarioGroup = Infections[I].ScenarioGroup then
      fBacterio.cobActual.Items.AddText(Infections[I].ScenarioActual);
  end;
  fBacterio.cobActual.ItemIndex := 0;
end;

{ Fill scenario special conditions combobox (for given scenario) }

procedure FillScenarios2(Scenario: string; var Infections: TInfections; var Suppls: TSuppls);

var
  N, I, J, P: Integer;
  S: string;
  Ins: Boolean;
  Scenarios2: array of string;

begin
  N := 0; SetLength(Scenarios2, N);
  for I := 0 to Length(Infections) - 1 do begin
    if Scenario = Infections[I].ScenarioActual then begin
      for J := 1 to 23 do begin
        if Infections[I].ScenarioBacterias[J] in ['1'..'9'] then begin
          // Search for special conditions for this scenario
          S := Suppls[StrToInt(Infections[I].ScenarioBacterias[J])].ScenarioUncommon;
          if (S <> '') and (S <> 'Uncommon') then begin
            // If there are named special conditions, store them into an arry
            P := Pos(',', S);
            if P = 0 then begin
              // There is one special condition stated
              N += 1; SetLength(Scenarios2, N);
              Scenarios2[N - 1] := S;
            end
            else begin
              // There are two special conditions stated
              N += 2; SetLength(Scenarios2, N);
              Scenarios2[N - 2] := Copy(S, 1, P - 1);
              Scenarios2[N - 1] := Copy(S, P + 2, Length(S));
            end;
          end;
        end;
      end;
    end;
  end;
  if Length(Scenarios2) = 0 then begin
    // If there are no special conditions, hide the combobox
    fBacterio.cobActual2.Visible := False;
  end
  else begin
    // If there are special conditions, fill them into combbox
    fBacterio.cobActual2.Items.Clear;
    fBacterio.cobActual2.Items.AddText(' ');                                   // add empty combobox item for general case
    for I := 0 to Length(Scenarios2) - 1 do begin
      Ins := True;
      // Eliminate doubles
      for J := 0 to fBacterio.cobActual2.Items.Count - 1 do begin
        if Scenarios2[I] = fBacterio.cobActual2.Items[J] then
          Ins := False;
      end;
      if Ins then
        fBacterio.cobActual2.Items.AddText(Scenarios2[I]);
    end;
    fBacterio.cobActual2.ItemIndex := 0;
    fBacterio.cobActual2.Visible := True;
  end;
end;

{************}
{ TfBacterio }
{************}

{ Application start: Initialisation }

procedure TfBacterio.FormCreate(Sender: TObject);

begin
  rgWall.Left := rgStain.Left;                                                 // Form controls re-arrangement
  ReadData(aBacterias, aInfections, aSuppls);                                  // Read program data
  FillScenarios(cobSelection.Text, aInfections);                               // Fill scenarios combobox
  FillScenarios2(cobActual.Text, aInfections, aSuppls);                        // Fill special conditions combobox
  iIdentification := 1;                                                        // Default identification by stain + shape
end;

{ Menu item "File > Exit": Exit application }

procedure TfBacterio.mFileExitClick(Sender: TObject);

begin
  Close;
end;

{ Menu items "Options > Identification > ...": Select identification method }

procedure TfBacterio.mOptionsId1Click(Sender: TObject);

// Identification by stain + shape

begin
  mOptionsId1.Checked := True; mOptionsId2.Checked := False;
  mOptionsId3.Checked := False; mOptionsId4.Checked := False;
  iIdentification := 1;
  rgStain.Visible := True; rgWall.Visible := False; rgGrowth.Visible := False;
end;

procedure TfBacterio.mOptionsId2Click(Sender: TObject);

// Identification by stain + shape + culture

begin
  mOptionsId1.Checked := False; mOptionsId2.Checked := True;
  mOptionsId3.Checked := False; mOptionsId4.Checked := False;
  iIdentification := 2;
  rgStain.Visible := True; rgWall.Visible := False; rgGrowth.Visible := True;
end;

procedure TfBacterio.mOptionsId3Click(Sender: TObject);

// Identification by cell wall + shape

begin
  mOptionsId1.Checked := False; mOptionsId2.Checked := False;
  mOptionsId3.Checked := True; mOptionsId4.Checked := False;
  iIdentification := 3;
  rgStain.Visible := False; rgWall.Visible := True; rgGrowth.Visible := False;
end;

procedure TfBacterio.mOptionsId4Click(Sender: TObject);

// Identification by cell wall + shape + culture

begin
  mOptionsId1.Checked := False; mOptionsId2.Checked := False;
  mOptionsId3.Checked := False; mOptionsId4.Checked := True;
  iIdentification := 4;
  rgStain.Visible := False; rgWall.Visible := True; rgGrowth.Visible := True;
end;

{ Menu item "Help > About": Dispaly application about }

procedure TfBacterio.mHelpAboutClick(Sender: TObject);

var
  S: string;

begin
  S := 'Basic bacterial identification.' + LineEnding + LineEnding;
  S += 'Version 1.0, © allu, December 2021.';
  MessageDlg('About "Bacteria"', S, mtInformation, [mbOK], 0);
end;

{ Button "Search" pushed: Display possible bacterial causes and search for most probable cause(s) }

procedure TfBacterio.btSearchClick(Sender: TObject);

var
  N, NC, NN, I, IX, J, P: Integer;
  Stain, Wall, Shape, Growth, TempS: string;
  S: AnsiString;
  OK: Boolean;
  CausesCommon, CausesUncommon: array of Integer;
  CausesNames: array of string;
  TempC: TBacteria;
  Causes: array of TBacteria;

begin
  // Get index for actual scenario (infection)
  for I := 0 to Length(aInfections) - 1 do begin
    if aInfections[I].ScenarioActual = cobActual.Text then
      IX := I;
  end;
  N := 0;
  // For each of the 23 bacteria, look up if they are a possible cause of this infection
  for I := 1 to 23 do begin
    if aInfections[IX].ScenarioBacterias[I] <> '-' then begin
      // Bacteria is a possible cause of this infection
      Inc(N); SetLength(Causes, N);
      SetLength(CausesCommon, N); SetLength(CausesUncommon, N); SetLength(CausesNames, N);
      Causes[N - 1] := aBacterias[I]; CausesCommon[N - 1] := 0; CausesUncommon[N - 1] := 0;
      // Find additional information for this infection-bacteria pair
      if aInfections[IX].ScenarioBacterias[I] in ['1'..'9'] then begin
        if aSuppls[StrToInt(aInfections[IX].ScenarioBacterias[I])].ScenarioUncommon <> '' then begin
          // Uncommon cause (or only under special conditions)
          CausesUncommon[N - 1] := StrToInt(aInfections[IX].ScenarioBacterias[I]);
        end;
        if aSuppls[StrToInt(aInfections[IX].ScenarioBacterias[I])].ScenarioBacteria <> '' then begin
          // More precise bacteria identification available
          CausesCommon[N - 1] := StrToInt(aInfections[IX].ScenarioBacterias[I]);
        end;
      end;
    end;
  end;
  NC := 0; NN := 0;
  // Create array with bacterial causes (bacteria names)
  for I := 0 to Length(Causes) - 1 do begin
    if (CausesUncommon[I] = 0) or (cobActual2.Text <> ' ') then begin
      CausesNames[I] := Causes[I].Name;                                        // common cause
      Inc(NC);
    end
    else begin
      CausesNames[I] := '{' + Causes[I].Name + '}';                            // uncommon cause (mark bacteria with "{}")
      Inc(NN);
    end;
    // If a more precise bacteria identification is possible, replace the "general" cause bacteria name by the more precise one(s)
    if CausesCommon[I] <> 0 then
      CausesNames[I] := StringReplace(CausesNames[I], aSuppls[CausesCommon[I]].ScenarioBacteria, aSuppls[CausesCommon[I]].ScenarioBacteriaDetails, []);
  end;
  // Sort bacteria names alphabetically
  for I := 1 to Length(CausesNames) - 1 do begin
    for J := 0 to I do begin
      if CausesNames[J] > CausesNames[I] then begin
        TempC := Causes[J]; Causes[J] := Causes[I]; Causes[I] := TempC;
        TempS := CausesNames[J]; CausesNames[J] := CausesNames[I]; CausesNames[I] := TempS;
      end;
    end;
  end;
  // Fill in the possible bacterial causes memo, grouping common and uncommon causes separately
  S := '';
  if NC > 0 then
    S := 'Common: ';
  for I := 0 to Length(CausesNames) - 1 do begin
    if I > 0 then
      S += ', ';
    S += CausesNames[I];
  end;
  if NC = 0 then
    S := 'Uncommon: ' + S
  else begin
    P := Pos(', {', S);
    if P <> 0 then
      S := StringReplace(S, ', {', LineEnding + 'Uncommon: ', []);
  end;
  S := StringReplace(S, '{', '', [rfReplaceAll]);
  S := StringReplace(S, '}', '', [rfReplaceAll]);
  edCauses.Lines.Clear;
  edCauses.Lines.AddText(S);
  // Determine most likely bacteria cause (if there is one for actual scenario and actual bacteria features)
  Stain := rgStain.Items[rgStain.ItemIndex];
  Wall := rgWall.Items[rgWall.ItemIndex];
  Shape := rgShape.Items[rgShape.ItemIndex];
  Growth := rgGrowth.Items[rgGrowth.ItemIndex];
  for I := 0 to Length(Causes) - 1 do begin
    OK := True;
    // Determination based on selected identification method
    case iIdentification of
      1: if (Causes[I].Stain <> Stain) or (Causes[I].Shape <> Shape) then
           OK := False;
      2: if (Causes[I].Stain <> Stain) or (Causes[I].Shape <> Shape) or (Causes[I].Growth <> Growth)then
           OK := False;
      3: if (Causes[I].Wall <> Wall) or (Causes[I].Shape <> Shape) then
           OK := False;
      4: if (Causes[I].Wall <> Wall) or (Causes[I].Shape <> Shape) or (Causes[I].Growth <> Growth)then
           OK := False;
    end;
    if not OK then
      CausesNames[I] := '';                                                    // set name of excluded bacteria to empty string
  end;
  // Fill in the most likely cause edit field
  S := '';
  for I := 0 to Length(CausesNames) - 1 do begin
    if CausesNames[I] <> '' then begin
      // Add all not excluded bacteria from the possible causes list to the most likely list
      if S <> '' then
        S += ', ';
      S += CausesNames[I];
    end;
  end;
  S := StringReplace(S, '{', '', [rfReplaceAll]);
  S := StringReplace(S, '}', '', [rfReplaceAll]);
  if S = '' then
    S := '???';                                                                // no bacteria with actual features causes this infection
  edCause.Text := S;
end;

{ Scenario group changes: Re-fill scenarios and special conditions comboboxes }

procedure TfBacterio.cobSelectionChange(Sender: TObject);

// Note: The special conditions combobox is not automatically refilled if the scenarios combobox is refilled
//       This is due to the fact that the TCombobox.OnChange method is only triggered by user done changes (not by program code)

begin
  FillScenarios(cobSelection.Text, aInfections);
  FillScenarios2(cobActual.Text, aInfections, aSuppls);
end;

{ Scenario changes: Re-fill special conditions comboboxe }

procedure TfBacterio.cobActualChange(Sender: TObject);

begin
  FillScenarios2(cobActual.Text, aInfections, aSuppls);
end;

end.

